# Copyright 2020 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

import time
import netq_netlink
import threading
import copy
from queue import Queue
from enum import Enum
from netq_lib.common.netq_pq import NetQPriorityQueue, QueueElem
from json import dumps as json_dumps, loads as json_loads

from netq_lib.orm import HOSTNAME
from netq_agent.cmd.netq_grpc_agent import *
from netq_lib.common.utils import is_sonic
from netq_agent.parsers.sonic.ifalias import update_ifalias_information

netlink_proc_obj = None
command_server_obj = None
global_working_sequence = {}

class AgentCommands(Enum):
    GET_FULL_STATE_ROUTE = 1
    GET_FULL_STATE_NETCONF = 2
    GET_FULL_STATE_NEIGHBOR = 3
    GET_FULL_STATE_ADDR = 4
    GET_FULL_STATE_MACFDB = 5
    GET_FULL_STATE_LINK = 6
    GET_FULL_STATE_VXLANREMOTEDEST = 7
    DUMP_NETLINK_MSG = 8


class NetlinkMsgType(Enum):
    NETLINK_MSG_TYPE_LINK = 'Link'
    NETLINK_MSG_TYPE_ROUTE = 'Route'
    NETLINK_MSG_TYPE_NEIGHBOR = 'Neighbor'
    NETLINK_MSG_TYPE_MACFDB = 'MacFdb'
    NETLINK_MSG_TYPE_ADDRESS = 'Address'
    NETLINK_MSG_TYPE_NETCONF = 'Netconf'
    NETLINK_MSG_TYPE_VXLANREMOTEDEST = 'VxlanRemoteDest'

msg_count_stats = dict()


class CommandServer(object):
    __MAX_LENGTH = 4096
    __SERVER = '/run/netq_agent_netlink'
    __DELIMITER = '$$$'

    def __init__(self, conf, logger):
        global command_server_obj
        self._conf = conf
        self._logger = logger
        self._ipc_q = Queue(512)
        command_server_obj = self

    def process_agent_command(self):
        try:
            if self._ipc_q.empty():
                return
            cmd = self._ipc_q.get()
            self.__process_agent_command(cmd)
        except Exception as e:
            return

    def __process_agent_command(self, command):
        command_args = command.split(",")
        command = eval(command_args[0])
        if (command == AgentCommands.GET_FULL_STATE_ROUTE):
            netq_netlink.get_full_state_route()
        elif (command == AgentCommands.GET_FULL_STATE_NEIGHBOR):
            netq_netlink.get_full_state_neigh()
        elif (command == AgentCommands.GET_FULL_STATE_LINK):
            netq_netlink.get_full_state_link()
        elif (command == AgentCommands.GET_FULL_STATE_ADDR):
            netq_netlink.get_full_state_addr()
        elif (command == AgentCommands.GET_FULL_STATE_NETCONF):
            netq_netlink.get_full_state_netconf()
        elif (command == AgentCommands.GET_FULL_STATE_MACFDB):
            netq_netlink.get_full_state_macfdb()
        elif (command == AgentCommands.GET_FULL_STATE_VXLANREMOTEDEST):
            netq_netlink.get_full_state_vxlanremotedest()
        elif (command == AgentCommands.DUMP_NETLINK_MSG):
            netq_netlink.dump_netlink_msg()
        else:
            self._logger.info("Unsupported Command {}".format(command))
                
    def send_command(self, command_info):
        cmd_str = command_info + self.__DELIMITER;
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            sock.connect(self.__SERVER)
        except Exception as e:
            self._logger.info("Failed to connect to Server {}".format(self.__SERVER))
            raise e
        self._logger.debug("Sending Command request for {}".format(cmd_str))
        sock.send(bytes(cmd_str, 'utf-8'))
        sock.close()
        return()

    def __process_socket_msg(self, clientsocket):

        try:
            buf = clientsocket.recv(self.__MAX_LENGTH).decode('utf-8')
            if buf == '':
                return
            self._logger.debug("Buffer recevied {}".format(buf))
            commands = buf.split(self.__DELIMITER)
            for cmd in commands:
                if cmd == '':
                   continue 
                self._logger.info("Processing command {}".format(cmd))
                self._ipc_q.put(cmd);
        except Exception as e:
            self._logger.info("Failed to process command {}".format(e))
        finally:
            self._logger.debug("Client Socket Disconnected")
            clientsocket.close()

    def start_server(self):
        try:
            os.unlink(self.__SERVER)
        except OSError:
            if os.path.exists(self.__SERVER):
                raise

        serversocket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        serversocket.bind(self.__SERVER)
        serversocket.listen(10)
        self._logger.info("Netlink Server Init Complete")

        while True:
            self._logger.debug("Waiting for netlink connections")
            (clientsocket, address) = serversocket.accept()
            self.__process_socket_msg(clientsocket)
 

class NetLinkProc(object):

    def __init__(self, conf, logger, pq, backpressure_queue, agent_pq):
        global netlink_proc_obj
        self._conf = conf
        self._channel = None
        self._logger = logger
        self._pq = pq
        self._q = Queue(250000)
        self._agent_pq = agent_pq
        self._bp_pq = backpressure_queue
        self._rt_cnt = 0
        self._grpc_object = NetQAgentGrpcClient(conf, logger, pq, backpressure_queue)
        self._wait_time = 0.0
        netlink_proc_obj = self
        self.__connectivity_state = False
        self.lock = threading.Lock()
        self.msg_count_stats = {}
        self.msg_count_stats["message_type_count"] = {}
        self.__netlink_update_time = 0.0
        self._NETLINK_UPDATE_TIME_SEC = 30.0
        self._GRPC_UPDATE_TIME = 0.5
        self._NETLINK_UPDATE_TIME = self._NETLINK_UPDATE_TIME_SEC / self._GRPC_UPDATE_TIME


    def __connect_reconnect_opta(self):
        while(True):
            if AGENTSTATUS.SUCCESS != self._grpc_object.reestablish_channel():
                self._logger.info("Failed to establish Connection with OPTA, retrying with backoff")
                if self._grpc_object.get_channel_retry_count() >= self._grpc_object.max_opta_retry:
                    self._grpc_object.reset_channel_retry_count()
                else:
                    self._grpc_object.update_channel_retry_count()
                timeVal = self._grpc_object.get_grpc_backoff_time()
                time.sleep(timeVal)
                continue
            self._logger.info("Starting Stream")
            break;
    
        self._opta_stub = NetqOPTAStreamServiceStub(self._grpc_object._channel)
        self.__connectivity_state = True


    def message_builder(self, msg_type, fullstate, json_out):

        seqno = 0;
        try:
            if (fullstate == TransMode.FIRST_SEGMENT) or \
                (fullstate == TransMode.PARTIAL_TRANSFORM):
                seqno = self._grpc_object.get_sequence_number()
                global_working_sequence[msg_type] = seqno
            else:
                seqno  = global_working_sequence[msg_type];
        except Exception as e:
            global_working_sequence[msg_type] = 0
        event_time = int(time.time() * 1000)


        if is_sonic() and NetlinkMsgType(msg_type) == NetlinkMsgType.NETLINK_MSG_TYPE_LINK:
            json_out = update_ifalias_information(json_out, self._logger)

        netq_message = NetqAgentMessage(aid=HOSTNAME,
                                        message_type=msg_type,
                                        seqno=seqno,
                                        ts=event_time,
                                        trans_mode=fullstate,
                                        type=0,
                                        message=json_out.encode('utf-8'))
        return netq_message


    def stream_libnl_msg(self, msg_type, json_out):
        try:
            rpc_msg = self.message_builder(msg_type, False, json_out)
            yield rpc_msg
            self._grpc_object._last_transmission_time = time.time()
        except:
            self._logger.error("GRPC  StreamMessage failed %s",  e)
            self._grpc_object._last_transmission_time = time.time()
            return

    def _is_json_string(self, json_out):
        try:
            json_dict = json.loads(json_out)
        except Exception as e:
            self._logger.error("Failed to load json string {}".format(json_out))
            return 0
        else:
            return 1


    def collect_data(self, msg_type, json_out, full_state):
        if self.__connectivity_state == False:
            self._logger.debug("OPTA Connection failure");
            return;
        if self._is_json_string(json_out):
            self._q.put((msg_type, full_state, json_out))
        else:
            self._logger.info("Failed to decode json object {}".format(json_out))
        if self._q.qsize() > 2000 or (full_state != TransMode.PARTIAL_TRANSFORM and self._q.qsize()>40):
            self.process_grpc_msg()

        '''Update Message Stats'''
        try:
            self.msg_count_stats['total_msg_count'] = self.msg_count_stats.get('total_msg_count', 0) + 1
            self.msg_count_stats['total_msg_bytes'] = self.msg_count_stats.get('total_msg_bytes', 0) + sys.getsizeof(json_out)
            if not full_state == TransMode.PARTIAL_TRANSFORM:
               self.msg_count_stats['full_state_count'] = \
                    self.msg_count_stats.get('full_state_count', 0) + 1
            else:
                self.msg_count_stats['partial_state_count'] = \
                    self.msg_count_stats.get('partial_state_count', 0) + 1
            self.msg_count_stats['netlink_msg_count'] = self.msg_count_stats.get('netlink_msg_count', 0) + 1
            
            msg_type_dict = self.msg_count_stats["message_type_count"] 
            msg_type_dict[msg_type] = msg_type_dict.get(msg_type, 0) + 1
            self.msg_count_stats["message_type_count"] = msg_type_dict
        except Exception as e:
            self._logger.error(" Caught Exception {}".format(e))


    def grpc_init_rate_control(self):
        try:
            self._wait_time = self._grpc_object._get_wait_time()
            self._logger.debug(
                "elapsed time {} wait time {} min interval {} queue size {}".format(self._grpc_object._elapsed_time, 
                                                                                    self._wait_time,
                                                                                    self._grpc_object._min_interval, self._q.qsize()))
            while self._wait_time > 0:
                q_len = self._q.qsize()
                if q_len > self._grpc_object._max_msg_threshold:
                    break
                else:
                    return False
                #self._grpc_object._wait_time = self._grpc_object._get_wait_time()

        except Exception as e:
            self._logger.info("Caught Wait time exception {}".format(e))
        return True

    def is_stream_available(self):
        if self.grpc_init_rate_control() and not self._q.empty():
            self._logger.debug("Stream data is available, attempting rpc")
            return True

        return False

    def get_stream(self):
        msg_cnt = 0
        try:
            if not self.grpc_init_rate_control():
                self._grpc_object._last_transmission_time = time.time()
                return 
            while True:
                try:
                    if self._q.empty():
                        return
                    (message_type, full_state, msg) = self._q.get()
                except Exception as e:
                    break
                rpc_msg = self.message_builder(message_type, full_state, msg)
                yield rpc_msg
                msg_cnt += 1
                if msg_cnt >= self._grpc_object._max_message_per_transaction:
                    self._logger.info("Reached max messages per transaction")
                    break
            self._grpc_object._last_transmission_time = time.time()
            return []

        except Exception as e:
            self._logger.error("GRPC %d StreamMessage failed %s", my_run_id, e)
            self._grpc_object._last_transmission_time = time.time()
            return []


    def process_netlink_stats(self):
        self.__netlink_update_time += self._GRPC_UPDATE_TIME
        if self.__netlink_update_time >= self._NETLINK_UPDATE_TIME_SEC:
            self.push_netlink_stats()
            self.__netlink_update_time = 0.0

    def process_grpc_msg(self):
        metadata = [('aid', HOSTNAME)]
        if self.__connectivity_state == False:
            self._logger.debug("Waiting for opta connection");
            return AGENTSTATUS.FAILURE
        while True:
            try:
                if not self.is_stream_available():
                    break
                with self.lock:
                    res, call = self._opta_stub.StreamMessage.with_call(self.get_stream(), metadata=metadata,
                                                                        timeout=self._grpc_object._grpc_call_timeout)
                    self._logger.debug("[{}] RPC Completed {} status {}".format(threading.current_thread().name, res, call.code()))
                if call.code() != grpc.StatusCode.OK:
                    self._logger.error("Agent Stream Failed {}".format(call.code()))
                    self.__connectivity_state = False
                    self._grpc_object._close()
                    return AGENTSTATUS.FAILURE
                break
            except grpc.RpcError as e:
                self._logger.error("Agent Stream Failed Code {} Info: {}".format(e.code(), e.details()))
                self.__connectivity_state = False
                self._grpc_object._close()
                return AGENTSTATUS.FAILURE
            except Exception as e:
                self._logger.error("Agent Stream failed error {}".format(e))
                self.__connectivity_state = False
                self._grpc_object._close()
                return AGENTSTATUS.FAILURE

    
    def run(self, config):
        self.__connect_reconnect_opta()
        time.sleep(10);
        netlink_thread = threading.Thread(target=netq_netlink.netlink_main, args=(HOSTNAME,))
        netlink_thread.start();
        cmd_srvr = CommandServer(config, self._logger)
        command_server = threading.Thread(target=cmd_srvr.start_server)
        command_server.start()
        while True:
            if (self._grpc_object.channel_state != ChannelState.CONNECTED) or  \
                    self.__connectivity_state == False:
                self.__connectivity_state = False
                self._logger.info("OPTA not reachable");
                self.__connect_reconnect_opta();
            self._logger.debug("Netlink active");
            if not command_server.is_alive():
                self._logger.info("Netlink Command server thread killed");
                break;
            if not netlink_thread.is_alive():
                self._logger.info("Netlink Parser thread killed");
                break;
            time.sleep(3);


        if netlink_thread:
            netlink_thread.join()
        if command_server:
            command_server.join()
        return

    def activate(self):
        priority = 2
        msg_type = "NETQ_RATE_CONTROL"
        pq_data = json_loads("[]")
        if self._pq is None:
            #self._logger.error("Netq Not in ready state")*/
            return 
        q_elem = QueueElem(priority, msg_type, False, pq_data)
        #self._logger.info("{}".format(libnl_msg))
        self._pq.putq(q_elem)


    def push_agent_info(self, msg_type, libnl_msg):
        priority = 1
        pq_data = json_loads(libnl_msg)
        if self._agent_pq is None:
            #self._logger.error("Netq Not in ready state")
            return
        q_elem = QueueElem(priority, msg_type, False, pq_data)
        self._agent_pq.putq(q_elem)

    def push_netlink_stats(self):
        priority = 1
        pq_data = json_dumps(self.msg_count_stats)
        #self._logger.info("Netlink Stats {}".format(pq_data))
        self.msg_count_stats = {}
        self.msg_count_stats["message_type_count"] = {}
        if self._agent_pq is None:
            #self._logger.error("Netq Not in ready state")
            return
        msg_type = "Stats"
        q_elem = QueueElem(priority, msg_type, False, pq_data)
        self._agent_pq.putq(q_elem)
    

    def log_netlink_msg(self, log_level, log_msg):
        if (log_level == "info"):
            self._logger.info(log_msg)
        elif (log_level == "debug"):
            self._logger.debug(log_msg)
        elif (log_level == "warning"):
            self._logger.warn(log_msg)
        else:
            self._logger.error(log_msg)


def proc_agent_info(msg_typ, libnl_msg):
    netlink_proc_obj.push_agent_info(msg_typ, libnl_msg)


def proc_libnl_msg(msg_type, libnl_msg, full_state):

    global netlink_proc_obj
    if netlink_proc_obj is None:
        print ("Netlink Object not ready")
        return 0

    netlink_proc_obj.collect_data(msg_type, libnl_msg, copy.copy(full_state))
    return 0


def proc_grpc_msg():
    command_server_obj.process_agent_command()
    netlink_proc_obj.process_grpc_msg()
    netlink_proc_obj.process_netlink_stats()

def proc_logger_msg(log_level, log_msg):
    netlink_proc_obj.log_netlink_msg(log_level, log_msg);

