# Copyright 2018, 2019, 2020, 2021 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

import time
import grpc
from multiprocessing.connection import wait
from netq_agent.interface import netq_stream_pb2_grpc, netq_stream_pb2, wjh_drop_aggregate_pb2
from netq_agent.cmd.grpc_client_base import GrpcClientBase
from netq_lib.common.enums import AGENTSTATUS, ModelConstants as MC
from netq_lib.orm import HOSTNAME

class GNMIDriver(GrpcClientBase):
    __MAX_GRPC_TIMEOUT = 30
    __MAX_GRPC_BACKOFF_TIME = 5

    def __init__(self, address, port, gnmi_pq, logger):
        super(GNMIDriver, self).__init__(
            logger, connection_callback=self.handle_gnmi_connection)
        self.target = '%s:%d' % (address, port)
        self._logger = logger
        self._gnmi_pq = gnmi_pq
        self.max_gnmi_retry = 3  # max wait for ready retires per channel
        self.__channel_retry_count = 0  # wait for ready state per channel retry count
        self.__channel_recreate_count = 0  # number of channels recreated count
        self._grpc_call_timeout = 300  # GRPC timeout
        self._logger.info("gNMI agent address: %s" % self.target)

    def handle_gnmi_connection(self):
        while self.is_active():
            if self._channel:
                self._channel.close()
            self._channel = grpc.insecure_channel(self.target)
            try:
                grpc.channel_ready_future(self._channel).result(
                    timeout=self.__MAX_GRPC_TIMEOUT)
                self._logger.info("gNMI channel connection successful")
                return AGENTSTATUS.SUCCESS
            except grpc.FutureTimeoutError:
                self._logger.info("gNMI channel Failed to connect")
                if self.__channel_retry_count >= self.max_gnmi_retry:
                    self.__channel_retry_count = 0
                    return AGENTSTATUS.FAILURE
                else:
                    self.__channel_retry_count = self.__channel_retry_count + 1
                time.sleep(self.__MAX_GRPC_BACKOFF_TIME)
        return AGENTSTATUS.SUCCESS

    def get_stream(self):
        pq = self._gnmi_pq
        fd_set = pq.get_reader_fds()
        try:
            epoll_input = wait(fd_set, 10)
            if not len(epoll_input):
                return
            msg_count = 0
            for queue_idx in range(0, pq.max_num_queue):
                for _, message_type, _, msg in pq.get_buffer(queue_idx):
                    if message_type in [MC.WJH_DROP_AGGREGATE,
                                        MC.WJH_L1_DROP_AGGREGATE]:
                        yield self._prepare_wjh_msg(message_type, msg.data)
                        msg_count += 1
                    else:
                        self._logger.error("Unsupported gNMI message "
                                           "received: %s", str(msg.data))
            if msg_count:
                self._logger.debug("gNMI StreamMessage sent %d "
                                   "messages successfully", msg_count)

        except Exception as e:
            self._logger.error("gNMI StreamMessage failed %s", e)
            return

    def run_contiguous_stream(self):
        sts = self.handle_connection()
        if sts != AGENTSTATUS.SUCCESS:
            self._logger.error(
                "failed to established gNMI Stream connection, "
                "max retries reached. exiting thread")
            return AGENTSTATUS.FAILURE

        gnmi_stub = netq_stream_pb2_grpc.NetqStreamStub(self._channel)
        metadata = [('aid', HOSTNAME)]
        while self.is_active():
            try:
                res, call = gnmi_stub.Stream.with_call(
                    self.get_stream(), metadata=metadata,
                    timeout=self._grpc_call_timeout)
                self._logger.debug(
                    "gNMI RPC Completed {} status {}".format(res, call.code()))
                if call.code() != grpc.StatusCode.OK:
                    self._logger.error("gNMI Stream Failed {}".format(
                        call.code()))
                    return
            except grpc.RpcError as e:
                self._logger.error(
                    "gNMI agent Stream Failed Code {} Info: {}".format(
                        e.code(), e.details()))
                return
            except KeyboardInterrupt:
                return
            except Exception as e:
                self._logger.error("Agent Stream failed error {}".format(e))
                return

    def _prepare_wjh_msg(self, msg_type, msg):
        if msg_type == MC.WJH_DROP_AGGREGATE:
            wjh_data = wjh_drop_aggregate_pb2.WjhDropAggregate(
                hostname=msg.get("hostname"),
                reason=msg.get("reason"),
                reason_id=msg.get("reason_id"),
                drop_type=msg.get("drop_type"),
                severity=msg.get("severity"),
                ingress_port=msg.get("ingress_port"),
                ingress_lag=msg.get("ingress_lag"),
                egress_port=msg.get("egress_port"),
                traffic_class=msg.get("traffic_class"),
                acl_rule_id=msg.get("acl_rule_id"),
                acl_bind_point=msg.get("acl_bind_point"),
                acl_name=msg.get("acl_name"),
                acl_rule=msg.get("acl_rule"),
                smac=msg.get("smac"),
                dmac=msg.get("dmac"),
                sip=msg.get("sip"),
                dip=msg.get("dip"),
                proto=msg.get("proto"),
                sport=msg.get("sport"),
                dport=msg.get("dport"),
                agg_count=msg.get("agg_count"),
                original_occupancy=msg.get("original_occupancy"),
                original_latency=msg.get("original_latency"),
                first_timestamp=int(msg.get("first_timestamp")),
                last_timestamp=int(msg.get("last_timestamp")),
                timestamp=int(msg.get("timestamp")))
            return netq_stream_pb2.NetqMessage(
                wjh_message=wjh_data)
        elif msg_type == MC.WJH_L1_DROP_AGGREGATE:
            wjh_data = wjh_drop_aggregate_pb2.WjhL1DropAggregate(
                hostname=msg.get("hostname"),
                ingress_port=msg.get("ingress_port"),
                is_port_up=msg.get("is_port_up"),
                port_down_reason=msg.get("port_down_reason"),
                description=msg.get("description"),
                state_change_count=msg.get("state_change_count"),
                symbol_error_count=msg.get("symbol_error_count"),
                crc_error_count=msg.get("crc_error_count"),
                first_timestamp=int(msg.get("first_timestamp")),
                last_timestamp=int(msg.get("last_timestamp")),
                timestamp=int(msg.get("timestamp")))
            return netq_stream_pb2.NetqMessage(
                wjh_l1_message=wjh_data)


if __name__ == '__main__':
    drop = {'acl_bind_point': 0, 'acl_name': '', 'acl_rule': '',
            'acl_rule_id': 0, '_active': True, 'agg_count': 1, 'deleted': False,
            'dip': '1.2.1.2', 'dmac': '00:bb:cc:11:22:33', 'dport': 80,
            'drop_type': 'L2', 'egress_port': '',
            'first_timestamp': 1612265393.981382, 'hostname': 'neo-switch01',
            'ingress_lag': '', 'ingress_port': 'swp8',
            'last_timestamp': 1612265393.981382,
            'message_type': 'WjhDropAggregate', 'original_latency': 0,
            'original_occupancy': 0, 'proto': 6,
            'reason': 'Source MAC equals destination MAC', 'severity': 'Error',
            'sip': '10.213.49.60', 'smac': '00:bb:cc:11:22:33', 'sport': 20,
            'timestamp': 1612265394.20512, 'traffic_class': 0, 'reason_id': 222}
    drop2 = {'acl_bind_point': 0, 'acl_name': '', 'acl_rule': '',
            'acl_rule_id': 0, '_active': True, 'agg_count': 1, 'deleted': False,
            'dip': '1.2.1.2', 'dmac': '00:bb:cc:11:22:22', 'dport': 80,
            'drop_type': 'Router', 'egress_port': '',
            'first_timestamp': 1612265393.981382, 'hostname': 'neo-switch01',
            'ingress_lag': '', 'ingress_port': 'swp8',
            'last_timestamp': 1612265393.981382,
            'message_type': 'WjhDropAggregate', 'original_latency': 0,
            'original_occupancy': 0, 'proto': 6,
            'reason': 'Ingress router interface is disabled', 'severity': 'Error',
            'sip': '10.213.49.100', 'smac': '00:bb:cc:11:22:22', 'sport': 20,
            'timestamp': 1612265394.20512, 'traffic_class': 0, 'reason_id': 240}
    try:
        print("start streaming")
        from netq_lib.common.netq_pq import create_gnmi_message_queues, QueueElem
        import logging
        import threading
        logger = logging.getLogger("gnmi")
        gnmi_pq = create_gnmi_message_queues(logger=logger)
        gnmi_driver = GNMIDriver("127.0.0.1", 7680, gnmi_pq, logger)
        gnmi_thread = threading.Thread(target=gnmi_driver.start_contiguous_stream)
        gnmi_thread.start()
        while True:
            for i in range(1024):
                drop['first_timestamp'] = time.time()
                drop['ingress_port'] = 'swp%d' % (int(i/5)+1)
                drop2['first_timestamp'] = time.time()
                drop2['ingress_port'] = 'swp%d' % (int(i/5)+1)
                gnmi_pq.putq(QueueElem(1, MC.WJH_DROP_AGGREGATE, False, drop))
                gnmi_pq.putq(QueueElem(1, MC.WJH_DROP_AGGREGATE, False, drop2))
            time.sleep(5)
    except KeyboardInterrupt:
        gnmi_driver._active = False
        gnmi_thread.join()

