# Copyright 2018, 2020 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

import os
import yaml
import ujson as json
from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import Roce, RoceConfig
from netq_lib.common.utils import save_fullstate, subprocess_wrapper, check_cl_version
from netq_lib.common.enums import Platform, ServiceStatus, DefaultPaths
from netq_agent.parsers.common.render import BaseRender
from netq_lib.common import configuration


class RoceRender(BaseRender):
    def __init__(self):
        super(RoceRender, self).__init__()
        self.port_id_name_mapping = dict()
        self.last_roce_config_modify_time = None

    def parse(self, cmd_output, svc_failed=False, **kwargs):
        fullstate = kwargs.get("fullstate")
        platform = kwargs.get('platform', '')
        if not configuration.is_cl4():
            return
        if Platform.MELLANOX not in platform:
            return

        if os.path.exists(DefaultPaths.ROCE_DUMMY_DATA_PATH):
            try:
                with open(DefaultPaths.ROCE_DUMMY_DATA_PATH, 'r') as f:
                    test_roce_data = yaml.load(f)

                    roce_intf_counters = test_roce_data.get('roce_counters', [])
                    for intf_counter in roce_intf_counters:
                        roce_obj = Roce()
                        roce_obj_attr = dir(roce_obj)
                        for ifname, intf_counters in list(intf_counter.items()):
                            setattr(roce_obj, 'ifname', ifname)
                            for key, value in list(intf_counters.items()):
                                if key in roce_obj_attr:
                                    setattr(roce_obj, key, value)
                        roce_obj.save()

            except Exception as e:
                pass
            return

        if not self.is_roce_enabled():
            self.delete_all_roce_counters()
            return

        pool_size_mapping = self.roce_pool_size()

        try:
            cmd_json_output = json.loads(cmd_output)
        except ValueError:
            self.delete_all_roce_counters()
            return ServiceStatus.FAIL

        fullstate_arr = []
        for per_intf_dict in list(cmd_json_output):
            roce_obj = Roce()
            roce_obj_attr = dir(roce_obj)
            for key, value in list(per_intf_dict.items()):
                if value is None:
                    continue
                if key in roce_obj_attr:
                    setattr(roce_obj, key, value)
                if key == 'port_id':
                    ifname = self._get_ifname_from_port_id(value)
                    setattr(roce_obj, 'ifname', ifname)

            for key, value in list(pool_size_mapping.items()):
                if key in roce_obj_attr:
                    setattr(roce_obj, key, value)

            fullstate_arr.append(roce_obj)
            if not fullstate:
                roce_obj.save()

        if fullstate:
            save_fullstate(fullstate_arr, Roce().__class__.__name__)

    def is_roce_enabled(self):
        roce_enabled = False
        if os.path.exists(DefaultPaths.ROCE_STATUS_FILE_PATH):
            try:
                with open(DefaultPaths.ROCE_STATUS_FILE_PATH, 'r') as fd:
                    for line in fd.readlines():
                        if 'traffic.roce_mode' in line:
                            mode = int(line.strip().split('=')[1].strip())
                            if mode != 0:
                                roce_enabled = True
                            break
            except Exception as e:
                pass
        return roce_enabled


    def is_roce_config_changed(self):
        roce_config_file_path = DefaultPaths.TRAFFIC_CONF_FILE_PATH
        if check_cl_version(major = 5):
            roce_config_file_path = DefaultPaths.QOS_CONF_FILE_PATH

        roce_config_changed = False
        if os.path.exists(roce_config_file_path):
            try:
                current_modify_time = int(os.path.getmtime(roce_config_file_path))
                if self.last_roce_config_modify_time is None:
                    roce_config_changed = True
                elif current_modify_time > self.last_roce_config_modify_time:
                    roce_config_changed = True

                if roce_config_changed:
                    self.last_roce_config_modify_time = current_modify_time
            except Exception as e:
                pass
        return roce_config_changed

    def parse_roce_config(self, cmd_output, svc_failed=False, **kwargs):
        fullstate = kwargs.get("fullstate")
        platform = kwargs.get('platform', '')
        if (Platform.MELLANOX not in platform) or (not configuration.is_cl4()):
            return

        if not self.is_roce_enabled():
            self.delete_all_roceconfig()
            return

        if not self.is_roce_config_changed() and not fullstate:
            return

        get_congestion_control = self.get_congestion_control()
        ets_priority_mode = self.get_ets_priority_mode()

        try:
            roce_status_output = subprocess_wrapper('/usr/lib/cumulus/mlxcmd --json roce status'.split())
            cmd_json_output = json.loads(roce_status_output)
        except ValueError:
            self.delete_all_roceconfig()
            return ServiceStatus.FAIL

        fullstate_arr = []
        for roce_status in list(cmd_json_output):
            roce_config_obj = RoceConfig()
            roce_config_obj.enabled_tc = get_congestion_control.get('enabled_tc', 'None')
            roce_config_obj.mode = get_congestion_control.get('mode', 'None')
            roce_config_obj.ecn_min_th  = get_congestion_control.get('min_th', 0)
            roce_config_obj.ecn_max_th  = get_congestion_control.get('max_th', 0)

            roce_config_obj.dscp_sp_map = roce_status.get('dscp',{})
            roce_config_obj.sp_pg_map = roce_status.get('pg_map',{})
            roce_config_obj.sp_tc_map = roce_status.get('tc_map',{})
            roce_config_obj.pfc_enabled_sp = roce_status.get('pfc_priority')
            roce_config_obj.roce_mode = roce_status.get('roce_mode')
            port_id = roce_status.get('port_id')
            roce_config_obj.ifname = self._get_ifname_from_port_id(port_id)

            roce_config_obj.pfc_rx_enabled = roce_status.get('pfc_enabled', {}).get('rx_enabled', 'no') == 'yes'
            roce_config_obj.pfc_tx_enabled = roce_status.get('pfc_enabled', {}).get('tx_enabled', 'no') == 'yes'
            roce_config_obj.ets_mode = ets_priority_mode

            if not fullstate:
                roce_config_obj.save()
            else:
                fullstate_arr.append(roce_config_obj)

        if fullstate:
            save_fullstate(fullstate_arr, RoceConfig().__class__.__name__)

    def delete_all_roceconfig(self):
        m_roceconfig = RoceConfig.memmodel()
        if m_roceconfig:
            for entry in m_roceconfig.query.filter(hostname=HOSTNAME):
                old_roce_config = RoceConfig()
                old_roce_config.copy(entry)
                old_roce_config.delete()

    def delete_all_roce_counters(self):
        m_roce = Roce.memmodel()
        if m_roce:
            for entry in m_roce.query.filter(hostname=HOSTNAME):
                old_roce = Roce()
                old_roce.copy(entry)
                old_roce.delete()

    def _get_ifname_from_port_id(self, port_id):
        if str(port_id) not in self.port_id_name_mapping:
            self.port_id_name_mapping = self.roce_port_info()
            print("Refreshed Port Id to Interface name mapping : {}".format(self.port_id_name_mapping))
        return self.port_id_name_mapping.get(str(port_id))
