# Copyright 2020 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt


import subprocess
from netq_lib.common.utils import subprocess_wrapper
from netq_agent.parsers.common.roce import RoceRender
from netq_lib.common.enums import ServiceStatus
import ujson as json

class CumulusRoceRender(RoceRender):
    def __init__(self):
        super(CumulusRoceRender, self).__init__()

    def roce_port_info(self):
        port_id_name_mapping = dict()
        try:
            mlx_port_info = subprocess_wrapper('/usr/lib/cumulus/mlxcmd --json ports info'.split())
            try:
                mlx_port_info_json = json.loads(mlx_port_info)
            except ValueError:
                return ServiceStatus.FAIL
            for port_id, port_status in list(mlx_port_info_json.items()):
                port_id_name_mapping[port_id] = port_status.get('ifname')
        except subprocess.CalledProcessError:
            pass
        return port_id_name_mapping


    def roce_pool_size(self):
        pool_size_mapping = dict()
        try:
            mlx_system_status = subprocess_wrapper('/usr/lib/cumulus/mlxcmd roce status --system'.split())
            for line in mlx_system_status.splitlines():
                split_line = line.split()
                if len(split_line) != 9:
                    continue
                pool_name, _, _, direction, _, pool_size, _, _, _ = line.split()
                pool_name_updated = pool_name.replace("-", "_") + '_size'
                pool_size_mapping[pool_name_updated] = pool_size
        except subprocess.CalledProcessError:
            pass
        return pool_size_mapping

    def get_congestion_control(self):
        congestion_control_map = dict()
        try:
            qos_redecn_info = subprocess_wrapper('/usr/lib/cumulus/mlxcmd --json qos redecn --port swp*'.split())
            try:
                qos_redecn_info_json_arr = json.loads(qos_redecn_info)
            except ValueError:
                return ServiceStatus.FAIL

            for qos_redecn_info_json in qos_redecn_info_json_arr:
                if 'config' in qos_redecn_info_json:
                    qos_configs = qos_redecn_info_json.get('config')
                    if qos_configs:
                        congestion_enabled_tc_lst =  list(qos_configs.keys())
                        congestion_enabled_tc = ','.join(congestion_enabled_tc_lst)
                        qos_config =  list(qos_configs.values())[0]
                        flow = list(qos_config.get('flow', {}).values())
                        if flow:
                            congestion_control_map['enabled_tc'] = congestion_enabled_tc
                            ecn_enabled = qos_config.get('ecn_enabled', 0) == 1
                            red_enabled = qos_config.get('red_enabled', 0) == 1

                            mode = 'None'
                            if ecn_enabled and not red_enabled:
                                mode = 'ECN'
                            elif not ecn_enabled and red_enabled:
                                mode = 'RED'
                            elif ecn_enabled and red_enabled:
                                mode = 'ECN+RED'

                            congestion_control_map['mode'] = mode
                            congestion_control_map['max_th'] =  flow[0].get('max', 0)
                            congestion_control_map['min_th'] =  flow[0].get('min', 0)
                            break

        except subprocess.CalledProcessError:
            pass

        return congestion_control_map


    def get_ets_priority_mode(self):
        dwrr_mode = ''
        try:
            roce_scheduler_info = subprocess_wrapper('/usr/lib/cumulus/mlxcmd --json scheduler show'.split())
            roce_scheduler_json = json.loads(roce_scheduler_info)
        except ValueError:
            return ServiceStatus.FAIL
        scheduler = roce_scheduler_json.get('list')

        if not scheduler:
            scheduler = roce_scheduler_json.get('scheduler-elements')

        if scheduler:
            roce_scheduler = next((item for item in list(scheduler) if (item.get('level') == 2 and item.get('index') == 3)), None)
            if roce_scheduler is not None:
                dwrr_mode = roce_scheduler.get('dwrr_mode')

        return dwrr_mode

def process_roce(cmd_output, svc_failed=False, **kwargs):
    roce_render.parse(cmd_output, svc_failed, **kwargs)

def process_roce_config(cmd_output, svc_failed=False, **kwargs):
    roce_render.parse_roce_config(cmd_output, svc_failed, **kwargs)

roce_render = CumulusRoceRender()
